omnibar .pan button {
  border: none;
  padding: 0;
  margin: 0;
  min-height: 12px;
  min-width: 14px;
  background: transparent;
  box-shadow: none;
  text-shadow: none;
  text-decoration: none;
  outline-offset: -1px;
}

omnibar .pan button:disabled {
  opacity: 0.45;
}

omnibar notification button {
  border: none;
  padding: 0;
  margin: 0;
  min-height: 24px;
  min-width: 24px;
  background: transparent;
  box-shadow: none;
  text-shadow: none;
  text-decoration: none;
  color: alpha(@theme_fg_color, 0.8);
}

omnibar entry {
  background-color: mix(@theme_bg_color, @content_view_bg, 0.25);
  color: @theme_fg_color;
}

omnibar:hover entry,
omnibar:active entry {
  background-color: @content_view_bg;
  color: @theme_fg_color;
}

omnibar:backdrop entry {
  background-color: @theme_bg_color;
}

/* Remove animation from linked buttons because it messes up the
 * joined state transition.
 */
omnibar > box.linked > button {
  transition-duration: 0;
}
