ideeditorsidebar > dzldockpaned:first-child stackswitcher {
  margin: 6px;
}
ideeditorsidebar .handle {
 border: 1px solid alpha(@borders, 0.3);
}
ideeditorsidebar label.title {
  margin: 8px 6px 3px 6px;
  font-weight: bold;
  font-size: 0.8em;
}
ideeditorsidebar list.open-pages row {
  padding: 1px 0;
  color: @theme_fg_color;
}
ideeditorsidebar list.open-pages row:selected {
  color: @theme_selected_fg_color;
}
ideeditorsidebar list.open-pages row:backdrop {
  color: @theme_unfocused_fg_color;
}
ideeditorsidebar list.open-pages row box > image:first-child {
  opacity: 0.55;
}
ideeditorsidebar list.open-pages row box > image:first-child {
  margin: 6px 8px;
  min-height: 16px;
  min-width: 16px;
}
ideeditorsidebar list.open-pages row box > button:last-child {
  background: none;
  box-shadow: none;
  border: none;
  padding: 0;
  margin: 0 9px 0 6px;
  color: @theme_fg_color;
  opacity: 0.55;
}
ideeditorsidebar list.open-pages row box > button:last-child:hover {
  opacity: 1;
}
ideeditorsidebar list.open-pages row box > button:last-child image {
  min-height: 16px;
  min-width: 16px;
}
ideeditorsidebar list.open-pages row box > button:last-child:backdrop {
  color: @theme_unfocused_fg_color;
}
ideeditorsidebar label.error {
  color: @error_color;
  font-weight: bold;
}
ideeditorsidebar label.error:backdrop {
  font-weight: normal;
}
ideeditorproperties button {
  padding: 2px 12px;
}
ideeditorproperties checkbutton {
  margin: 8px 0 0 0;
  outline-offset: 2px;
  padding: 0;
}
ideeditorproperties checkbutton check {
  margin: 0 8px 0 0;
}
ideeditorproperties treeview {
  color: @theme_fg_color;
}
ideeditorproperties treeview:selected:backdrop,
ideeditorproperties treeview:selected {
  color: @theme_selected_fg_color;
}
ideeditorproperties treeview:backdrop {
  color: @theme_unfocused_fg_color;
}
ideeditorproperties button.control.flat {
  padding: 0;
  margin: 0 8px 0 0;
  min-width: 16px;
  min-height: 16px;
}
ideeditorsidebar .flat-menu-button {
  border: none;
  background: transparent;
  box-shadow: none;
  opacity: 0.5;
  padding: 0;
  margin: 0;
  outline-offset: -3px;
}
ideeditorsidebar .flat-menu-button:checked,
ideeditorsidebar .flat-menu-button:hover {
  opacity: 1;
}

/* utilities panel buttons */
ideeditorutilities dzltab {
  background: none;
  background-color: @theme_bg_color;
  background-image: none;
  background-size: 6px 6px, 6px 6px;
  background-repeat: no-repeat;
  background-position: 3px 3px;
  padding: 6px 8px;
  margin: 0;
  border-style: solid;
  border-color: @borders;
  border-width: 1px 1px 0 1px;
}
ideeditorutilities dzltab:checked {
  background-color: mix(@theme_bg_color, @borders, 0.3);
}
ideeditorutilities dzltab:hover {
  background-color: mix(@theme_bg_color, @borders, 0.1);
}
ideeditorutilities dzltab:first-child {
  border-radius: 3px 3px 0 0;
  border-width: 1px 1px 0 1px;
}
ideeditorutilities dzltab:last-child {
  border-radius: 0 0 3px 3px;
  border-bottom-width: 1px;
}

ideeditorutilities dzltab.needs-attention {
  background-image: -gtk-gradient(radial, center center, 0, center center, 0.5, to(@theme_selected_bg_color), to(transparent));
  background-size: 6px 6px, 6px 6px;
  background-repeat: no-repeat;
  background-position: 3px 3px;
}
ideeditorutilities dzltabstrip {
  margin: 5px;
  border-style: none;
}

/* hide extra textview cursors for non-visible views:
 * https://gitlab.gnome.org/GNOME/gnome-builder/issues/999
 */
textview:not(:focus) text {
  caret-color: transparent;
  -gtk-secondary-caret-color: transparent;
}


vte-terminal {
  background-color: @content_view_bg;
  color: @theme_fg_color;
}
