@import url("resource:///plugins/glade/themes/Adwaita-shared.css");

/* Draw our grid over the glade background pattern */
gbpgladeview viewport {
  background-color: #f6f7f8;
  background-size: 8px 8px;
  background-image: repeating-linear-gradient(0deg, rgba(.125,.125,.125,.025), rgba(.125,.125,.125,.025) 1px, transparent 1px, transparent 8px),
                    repeating-linear-gradient(-90deg, rgba(.125,.125,.125,.025), rgba(.125,.125,.125,.025) 1px, transparent 1px, transparent 8px);
}
.glade-chooser {
  background-color: #f6f7f8;
}
